

def castValueForDataType(datatype: str, value):
    """ Cast rdf literal into pythonic type """
    if datatype in [
        "http://www.w3.org/2001/XMLSchema#float",
        "http://www.w3.org/2001/XMLSchema#double",
        "http://www.w3.org/2001/XMLSchema#timestamp"
    ]:
        return float(value)
    elif datatype in [
        "http://www.w3.org/2001/XMLSchema#int",
        "http://www.w3.org/2001/XMLSchema#integer"
    ]:
        return int(value)
    elif datatype in [
        "http://www.w3.org/2001/XMLSchema#boolean"
    ]:
        return value if type(value) == bool else value == "true"
    elif datatype in [
        "http://www.w3.org/2001/XMLSchema#date",
        "http://www.w3.org/2001/XMLSchema#dateTime",
        "http://www.w3.org/2001/XMLSchema#dateTimeStamp"
    ]:
        return int(value)
    else:
        return value


def parse_construct_result(bindings):
    """ Parse result of construct query to convert bindings into a set of entities with corresponding properties """
    entities = []
    for binding in bindings:
        rdf_props = filter(lambda x: x not in ['@id', '@type'], binding.keys())
        entity_props = []
        for rdf_prop in rdf_props:
            value = binding[rdf_prop]
            if '#' in rdf_prop:
                prefix, prop_name = rdf_prop.split('#')
            else:
                prefix = rdf_prop[0:rdf_prop.rfind('/')+1]
                prop_name = rdf_prop[rdf_prop.rfind('/')+1:]
            entity_props.append([prefix, prop_name, value])
        entities.append({
            '@id': binding['@id'],
            '@type': binding['@type'],
            'properties': entity_props
        })

    parsed_objects = []
    for focused_entity in entities:
        fields = {
            "id": focused_entity['@id'],
            "uri": focused_entity['@id'],
            "types": focused_entity['@type']
        }
        for prefix, prop_name, value in focused_entity['properties']:
            if '@id' in value[0]:
                fields[prop_name] = []
                for target in value:
                    nested_entity = next((obj for obj in entities if obj['@id'] == target['@id']), None)
                    fields[prop_name].append(target['@id'])
            elif '@language' in value[0]:
                fields[prop_name] = list(map(lambda label: label['@value'], value))
                fields[f'{prop_name}_locales'] = list(map(lambda label: label.get('@language', 'en'), value))
            else:
                fields[prop_name] = castValueForDataType(value[0].get('@type', None), value[0]['@value'])
        parsed_objects.append(fields)

    return parsed_objects



import uuid
import datetime
import logging
import json
from ..toolkit.publisher import AbstractPublisher
from .. import AMQPClient


class GraphPublisher(AbstractPublisher):
    """ Extends publisher to serve methods adapted to graph queries """

    def __init__(self, amqp_driver: AMQPClient, sender_id: str, repository_name: str):
        super(GraphPublisher, self).__init__(amqp_driver, sender_id)
        self._repository_name = repository_name

    async def ask(self, query: str):
        """ Executes ASK query in parameter """
        if 'ASK' not in query:
            raise Exception(f'Use ask select method for doing a select query')
        return await self.publish('graph.ask', query, {
            "format": "SPARQL",
            "repositories": [self._repository_name]
        })

    async def select(self, query: str):
        """ Executes SELECT query in parameter """
        if 'SELECT' not in query:
            raise Exception(f'Use publisher construct method for doing a construct query')
        return await self.publish('graph.select', query, {
            "format": "SPARQL",
            "repositories": [self._repository_name]
        })

    async def construct(self, query: str):
        """ Executes CONSTRUCT query in parameter """
        if 'CONSTRUCT' not in query:
            raise Exception(f'Use publisher select method for doing a select query')
        return await self.publish('graph.construct', query, {
            "format": "SPARQL",
            "repositories": [self._repository_name]
        })



import json
import os
import datetime
from .publisher import GraphPublisher
from .parser import parse_construct_result
from .sparql import SPARQLBuilder


class GraphDriver(object):
    """ Driver for triplestore database """

    def __init__(self, amqp_client, sender_id, repository_name, prefixes_mapping, nodes_namespace_uri, nodes_prefix):
        self._publisher = GraphPublisher(amqp_client, sender_id, repository_name)
        self._sparql_builder = SPARQLBuilder(prefixes_mapping, nodes_namespace_uri, nodes_prefix)

    def get_graph_publisher(self):
        return self._publisher

    async def raw_select(self, query: str):
        """ Executes a raw select query
            Is equivalent to get publisher from this object and call its select method
        """
        result = await self._publisher.select(query)
        return result

    async def raw_construct(self, query: str, parse_result=True):
        """ Executes a raw construct query
            if parse_result is True, parse bindings result and convert into properties grouped by entity returned
        """
        bindings = await self._publisher.construct(query)
        if parse_result:
            objects = parse_construct_result(bindings)
        else:
            objects = bindings
        return objects

    async def get_nodes(self, node_type: str, object_ids: [str] = None):
        """ Get nodes and their properties corresponding to node_type """
        sparql_query = self._sparql_builder.build_construct_objects_query(node_type)
        bindings = await self._publisher.construct(sparql_query)
        objects = parse_construct_result(bindings)
        return objects

    async def get_node(self, node_type: str, object_id: str):
        """ Get node and its properties for node_type and object_id """
        sparql_query = self._sparql_builder.build_construct_object_query(node_type, object_id)
        bindings = await self._publisher.construct(sparql_query)
        objects = parse_construct_result(bindings)
        return objects[0] if len(objects) > 0 else None



import uuid
import json
import logging
import traceback
from typing import Callable
from functools import partial
from aio_pika import connect_robust, IncomingMessage, Message, Exchange, ExchangeType


LOGGER = logging.getLogger(__name__)


class AMQPClient(object):
    """AMQP client for connecting a rabbitMQ bus"""

    def __init__(self, amqp_url: str, exchange_name: str, loop, options={}):
        self._url = amqp_url
        self._exchange_name = exchange_name
        self._loop = loop
        self._prefetch_count = options.get('prefetch_count', 10)
        self._connection = None
        self._channel = None
        self._futures = {}
        self._consumer_id = str(uuid.uuid4())
        self._listening_queue_number = 0

        self._rpc_callback_queue = None

    async def connect(self):
        """ Connect to the amqp bus """
        self._connection = await connect_robust(self._url, loop=self._loop)
        self._channel = await self._connection.channel()
        await self._channel.set_qos(prefetch_count=1)

        await self._channel.declare_exchange(self._exchange_name, ExchangeType.TOPIC)

        rpc_queue_name = f"{self._exchange_name}_{self._consumer_id}_rpc_callback_queue"
        self._rpc_callback_queue = await self._channel.declare_queue(rpc_queue_name,
                                                                     exclusive=True,
                                                                     auto_delete=True,
                                                                     durable=False)

        await self._rpc_callback_queue.consume(self._on_rpc_response, no_ack=True)

        return self

    async def publish(self, command: str, body: object, options={}):
        """ Publish a message to the bus """
        if self._channel is None:
            raise Exception('AMQP client is not connected')

        exchange = await self._channel.get_exchange(self._exchange_name)  \
            if self._exchange_name \
            else self._channel.default_exchange

        await exchange.publish(
            Message(
                json.dumps(body).encode(),
                content_type=options.get("content_type", "application/json"),
                correlation_id=options.get("correlation_id", None),
                reply_to=options.get("reply_to", None)
            ),
            routing_key=command,
            mandatory=options.get("mandatory", True)
        )

    async def request(self, command: str, body: object):
        """ Trigger a RPC call with command value as routing_key, and wait for result before return """
        correlation_id = str(uuid.uuid4())
        future = self._loop.create_future()

        self._futures[correlation_id] = future

        await self.publish(command, body, {
            "content_type": "application/json",
            "correlation_id": correlation_id,
            "reply_to": self._rpc_callback_queue.name
        })

        return await future

    async def listen(self, routing_key: str, callback: Callable[[object], object], queue_name=None):
        """ Listen for messages passing on queue_name with routing_key, and executes callback
            if queue_name is not defined, a custom and unique queue_name is used
        """
        if self._channel is None:
            raise Exception('AMQP client is not connected')

        routing_key = routing_key or '#'
        self._listening_queue_number = self._listening_queue_number + 1
        queue_name = queue_name or f"{self._exchange_name}_{self._consumer_id}_{self._listening_queue_number}:{routing_key}"

        queue = await self._channel.declare_queue(queue_name, exclusive=True, auto_delete=True, durable=False)

        listen_exchange = await self._channel.get_exchange(self._exchange_name)  \
            if self._exchange_name \
            else self._channel.default_exchange
        await queue.bind(listen_exchange, routing_key=routing_key)

        await queue.consume(partial(self._on_listen_callback, self._channel.default_exchange, callback))

    def _on_rpc_response(self, message: IncomingMessage):
        """ RPC response
            Future linked to RPC call got message result, and so we can keep running code with response
        """
        try:
            future = self._futures.pop(message.correlation_id)
            future.set_result(message.body.decode())
        except Exception as e:
            print(message)

    async def _on_listen_callback(self, exchange: Exchange, callback: Callable[[object], object], message: IncomingMessage):
        """ Listen response wrapper
            This wrapper process message through callback, and send back response if reply_to is defined in message
        """
        with message.process():
            body = json.loads(message.body.decode())
            try:
                result = await callback(body)
                response = {
                    "headers": {"status": "OK "},
                    "body": result
                }
            except Exception as e:
                LOGGER.error(f'Error "{str(e)}" occurred on message : {message}')
                LOGGER.error(traceback.format_exc())
                response = {
                    "headers": {"status": "ERROR"},
                    "body": {"errorMessage": f'An error occurred on processing message, see logs...'}
                }

            if message.reply_to:
                await exchange.publish(
                    Message(
                        content_type="application/json",
                        body=json.dumps(response).encode(),
                        correlation_id=message.correlation_id
                    ),
                    routing_key=message.reply_to,
                )



import json
from ..drivers import GraphDriver, MongoDriver, IndexDriver
from .model import Model
from .keys import DATATYPE_MAPPER, KEY_NAME, KEY_RDF_TYPE, KEY_INDEX_TYPE,\
    KEY_MONGO_TYPE, KEY_DATATYPE, KEY_RELATED_TYPE, KEY_NESTED, KEY_PLURAL


class Datastore(object):
    """ Datastore for uniformizing query to graph and mongo databases """

    _schema_types: object
    _prefixes_mapping: object
    _nodes_namespace_uri: str
    _nodes_prefix: str
    _graph_driver: GraphDriver
    _mongo_driver: MongoDriver
    _index_driver: IndexDriver
    _model_class_mapper: object

    def __init__(
        self,
        json_schema: str,
        prefixes_mapping: object,
        nodes_namespace_uri: str,
        nodes_prefix: str,
        graph_driver: GraphDriver = None,
        mongo_driver: MongoDriver = None,
        index_driver: IndexDriver = None,
        model_class_mapper: object = None,
    ):
        self._schema_types = json_schema
        self._prefixes_mapping = prefixes_mapping
        self._nodes_namespace_uri = nodes_namespace_uri
        self._nodes_prefix = nodes_prefix
        self._graph_driver = graph_driver
        self._mongo_driver = mongo_driver
        self._index_driver = index_driver
        self._model_class_mapper = model_class_mapper

    def get_mongo_driver(self):
        return self._mongo_driver

    def get_graph_driver(self):
        return self._graph_driver

    def get_index_driver(self):
        return self._index_driver

    def close_connections(self):
        """ Close driver connections """
        self._mongo_driver.close()

    def get_schema_type(self, typename: str):
        """ Return schema corresponding to typename """
        return next((schema for schema in self._schema_types.values() if schema[KEY_NAME] == typename), None)

    async def get_objects(self, typename: str, object_ids: [str] = None, use_graph: bool = False):
        """ Return all objects of type defined in argument
            Results are merged objects of triplestore and mongo databases
        """
        schema_spec = self.get_schema_type(typename)
        if not schema_spec:
            raise Exception(f'Unknow type {typename}')

        graph_type = schema_spec.get(KEY_RDF_TYPE, None)
        index_type = schema_spec.get(KEY_INDEX_TYPE, None)
        mongo_type = schema_spec.get(KEY_MONGO_TYPE, None)

        graph_objects = []
        mongo_objects = []

        if not use_graph and self._index_driver and index_type:
            document_ids = list(map(lambda x: self.to_absolute_uri(x), object_ids)) if object_ids else None
            graph_objects = await self._index_driver.get_nodes(index_type, document_ids)
            graph_objects = [{**obj, "id": self.to_prefixed_uri(obj["id"])} for obj in graph_objects]
        elif self._graph_driver and graph_type:
            if object_ids:
                raise Exception('Query GraphIndex.getNodes with objects_ids is not yet supported...')
            graph_objects = await self._graph_driver.get_nodes(graph_type)
            graph_objects = [{**obj, "id": self.to_prefixed_uri(obj["id"])} for obj in graph_objects]

        if self._mongo_driver and mongo_type:
            mongo_ids = list(map(lambda x: self.to_prefixed_uri(x), object_ids)) if object_ids else None
            mongo_objects = self._mongo_driver.find_documents(mongo_type, mongo_ids)

        merged_objects = []
        mongo_objects_id = list(map(lambda x: x['id'], mongo_objects))
        if graph_type:
            for graph_object in graph_objects:
                matching_mongo_object = next((obj for obj in mongo_objects if obj['id'] == graph_object['id']), None) or {}
                merged_objects.append({
                    **graph_object,
                    **matching_mongo_object,
                    "type": typename
                })
        else:
            merged_objects = mongo_objects

        return merged_objects

    async def get_object(self, typename: str, object_id: str, use_graph=False):
        """ Return object corresponding to id and type defined in argument
            Result is merge of object found in triplestore and mongo databases if any
        """
        schema_spec = self.get_schema_type(typename)
        if not schema_spec:
            raise Exception(f'Unknow type {typename}')

        graph_type = schema_spec.get(KEY_RDF_TYPE, None)
        index_type = schema_spec.get(KEY_INDEX_TYPE, None)
        mongo_type = schema_spec.get(KEY_MONGO_TYPE, None)

        graph_object = {}
        mongo_object = {}

        if not use_graph and self._index_driver and index_type:
            graph_object = await self._index_driver.get_node(index_type, self.to_absolute_uri(object_id))
            graph_object = {**graph_object, "id": self.to_prefixed_uri(graph_object['id'])} if graph_object else None
        elif self._graph_driver and graph_type:
            graph_object = await self._graph_driver.get_node(graph_type, self.to_absolute_uri(object_id))
            graph_object = {**graph_object, "id": self.to_prefixed_uri(graph_object['id'])} if graph_object else None

        if self._mongo_driver and mongo_type:
            mongo_object = self._mongo_driver.find_document(mongo_type, self.to_prefixed_uri(object_id)) or None

        if not graph_object and not mongo_object:
            return None

        merged_object = {
            **(graph_object or {}),
            **(mongo_object or {}),
            "type": typename
        }

        return merged_object

    def to_absolute_uri(self, uri: str):
        """ Convert an uri into absolute uri """
        if uri.startswith('http://'):
            return uri
        elif self._nodes_prefix in uri:
            return uri.replace(f'{self._nodes_prefix}:', f'{self._nodes_namespace_uri}/')
        else:
            matching_prefix = next((prefix for prefix in self._prefixes_mapping.keys() if prefix in uri), None)
            if matching_prefix:
                return uri.replace(f'{matching_prefix}:', self._prefixes_mapping[matching_prefix])
        # Last case: no prefix at all, we put nodes namespace uri
        return f'{self._nodes_namespace_uri}/{uri}'

    def to_prefixed_uri(self, uri: str):
        """ Convert an uri into prefixed uri """
        if f'{self._nodes_prefix}:' in uri:
            return uri
        elif ':' not in uri and '#' not in uri:
            return f'{self._nodes_prefix}:{uri}'
        elif self._nodes_namespace_uri in uri:
            return uri.replace(f'{self._nodes_namespace_uri}/', f'{self._nodes_prefix}:')
        else:
            prefix, ns_uri = next(
                ((prefix, ns_uri) for prefix, ns_uri in self._prefixes_mapping.items() if ns_uri in uri), (None, None)
            )
            if ns_uri:
                if '#' in uri:
                    return uri.replace(f'{ns_uri}#', f'{prefix}:')
                else:
                    return uri.replace(f'{ns_uri}/', f'{prefix}:')

        # Last case: no prefix at all, we put nodes prefix
        return f'{self._nodes_prefix}:{uri}'

    def to_id(self, uri: str):
        """ Convert an uri into id """
        uri = uri.replace(f'{self._nodes_namespace_uri}/', '')
        uri = uri.replace(f'{self._nodes_prefix}:', '')
        return uri

    def to_model(self, typename, obj):
        """ Instanciate object as a class matching typename in argument
            Raise exception if model_class_mapper was not defined in datastore
            Raise as well exception if model class not found
        """
        if not self._model_class_mapper:
            raise Exception("Cannot call 'Datastore.to_model' function as no class mapper was passed to Datastore")

        schema_type = self.get_schema_type(typename)
        model_cls = self._model_class_mapper.get(typename, Model)
        if not model_cls:
            raise Exception(f"No model class defined for object of type {typename}")

        return model_cls(obj, schema_type, self)



from .keys import DATATYPE_MAPPER, KEY_NAME, KEY_RDF_TYPE, KEY_INDEX_TYPE,\
    KEY_MONGO_TYPE, KEY_DATATYPE, KEY_RELATED_TYPE, KEY_NESTED, KEY_PLURAL, \
    KEY_INTERNAL_PROP


class Model(object):
    """ Model base class for all objects queried from graph """

    _internal_props: object
    _schema: object
    _datastore: object

    def __init__(self, props, schema, datastore):
        self._internal_props = props
        self._schema = schema
        self._datastore = datastore

    @classmethod
    async def load(cls, object_id, datastore):
        """ Query an object through datastore and instanciate corresponding model """
        schema_type = cls.get_schema_type()
        graph_object = await datastore.get_object(schema_type, object_id)
        if not graph_object:
            return None
        return datastore.to_model(schema_type, graph_object)

    @property
    def id(self):
        return self.get_prop('id')

    def get_prop(self, name):
        """ Get prop from internal props if existing else None """
        prop_spec = self._schema.get("props", {}).get(name, None)
        if not prop_spec:
            raise Exception(f'Property {name} not in schema : ${self._schema}')

        target_datatype = prop_spec[KEY_DATATYPE]
        is_plural = prop_spec[KEY_PLURAL]
        if target_datatype not in DATATYPE_MAPPER:
            raise Exception(f'Datatype {target_datatype} not in datatype mapper')

        convert_to_type = DATATYPE_MAPPER[target_datatype]
        prop_value = self._internal_props.get(name, None)
        if not prop_value:
            return [] if is_plural else None

        if not is_plural:
            value = prop_value[0] if isinstance(prop_value, list) else prop_value
            return convert_to_type(value)

        return [convert_to_type(v) for v in prop_value]

    async def get_link(self, name):
        """ Get prop from internal props if existing, and query datastore for getting matching objects with ids """
        links = self._schema.get("links", {})
        link_spec = links.get(name, None)
        if not link_spec:
            link_spec = next((spec for spec in links.values() if spec[KEY_INTERNAL_PROP] == name), None)
        if not link_spec:
            raise Exception(f'Property {name} not in schema : ${self._schema}')

        related_type = link_spec[KEY_RELATED_TYPE]
        is_plural = link_spec[KEY_PLURAL]
        is_nested = link_spec[KEY_NESTED]

        targets = self._internal_props.get(name, None)
        if not targets or len(targets) == 0:
            return [] if is_plural else None

        targets = [targets] if not isinstance(targets, list) else targets

        if not is_nested:
            targets = await self._datastore.get_objects(related_type, targets)

        targets = [self._datastore.to_model(related_type, target) for target in targets]
        return targets if is_plural else targets[0] if len(targets) > 0 else None

    def to_json(self):
        """ Return model object as json """
        return {**self._internal_props}



import asyncio
from elasticsearch import Elasticsearch


class IndexDriver(object):
    """ Driver for elasticsearch index """

    def __init__(self, host, port, user, pwd, index_prefix):
        self._client = Elasticsearch(
            {host: host},
            port=port,
            http_auth=(user, pwd)
        )
        self._prefix = index_prefix

    async def get_nodes(self, index_type: str, object_ids: [str] = None):
        if object_ids:
            query = {
                "query": {
                    "ids": {
                        "values": object_ids
                    }
                }
            }
        else:
            query = {
                "query": {
                    "match_all": {}
                }
            }
        results = self._client.search(
            index=self._build_index(index_type),
            body=query,
            size=20
        )
        hits = results['hits']['hits']
        return [{**hit['_source'], "id": hit["_id"]} for hit in hits]

    async def get_node(self, index_type: str, object_id: str):
        result = self._client.get(
            index=self._build_index(index_type),
            id=object_id
        )
        return {**result['_source'], "id": result["_id"]}

    def _build_index(self, index_type):
        return f'{self._prefix}-{index_type}'



import json
import uuid
import logging
import datetime


LOGGER = logging.getLogger(__name__)


class AbstractPublisher(object):
    """ Base class for a amqp message publisher """

    def __init__(self, amqp_driver, sender_id: str):
        self._amqp_driver = amqp_driver
        self._sender_id = sender_id

    async def publish(self, command, body, options={}, expect_response=True):
        """ Publish a message through amqp bus
            if expect_response is True, publish a RPC message and await for response
        """

        # Synaptix format for querying graph or index, perhaps allow other format ?
        message = {
            "headers": {
                "command": command,
                "sender": self._sender_id or uuid.uuid4(),
                "date": datetime.datetime.now().timestamp(),
                **options
            },
            "body": body
        }

        if expect_response:
            response = await self._amqp_driver.request(command, message)

            response = json.loads(response) if type(response) == str else response

            if response["headers"]["status"] == 'OK':
                return response["body"]
            elif response["headers"]["status"] == 'ERROR':
                LOGGER.error(f'An error occurred on processing message : {body}')
                raise Exception(f'Error on message : {response["body"]["errorMessage"]}')

            raise Exception(f'Response with unhandled status : {response}')
        else:
            await self._amqp_driver.publish(command, message)




DATATYPE_MAPPER = {
    # Rdf types
    'rdfs:Literal': str,
    'http://www.w3.org/2000/01/rdf-schema#Literal': str,
    'http://www.w3.org/2001/XMLSchema#Literal': str,
    'http://www.w3.org/2001/XMLSchema#string': str,
    'http://www.w3.org/2001/XMLSchema#anyURI': str,
    'http://www.w3.org/2001/XMLSchema#short': int,
    'http://www.w3.org/2001/XMLSchema#int': int,
    'http://www.w3.org/2001/XMLSchema#integer': int,
    'http://www.w3.org/2001/XMLSchema#float': float,
    'http://www.w3.org/2001/XMLSchema#boolean': bool,
    'http://www.w3.org/2001/XMLSchema#date': str,
    'http://www.w3.org/2001/XMLSchema#dateTime': str,
    'http://www.w3.org/2001/XMLSchema#dateTimeStamp': str,
    'http://www.w3.org/2001/XMLSchema#duration': str,
    'http://www.w3.org/2001/XMLSchema#gMonth': str,
    'http://www.opengis.net/ont/geosparql#wktLiteral': str,
    'http://www.opengis.net/ont/geosparql#gmlLiteral': str,
    'http://www.elzeard.co/ontologies/c3po/time#rdate': str,
    # Mongo types
    "Int": int,
    "Float": float,
    'String': str,
    'Boolean': bool
}

KEY_NAME = 'name'
KEY_RDF_TYPE = 'rdfType'
KEY_INDEX_TYPE = 'indexType'
KEY_MONGO_TYPE = 'mongoType'
KEY_DATATYPE = 'dataType'
KEY_INTERNAL_PROP = 'internalProp'
KEY_RELATED_TYPE = 'relatedType'
KEY_NESTED = 'nested'
KEY_PLURAL = 'plural'



from typing import TypedDict
from pymongo import MongoClient


class MongoDriver(object):
    """ Driver for mongo database """

    def __init__(self, url, db_name):
        self._db_name = db_name
        self._mongo_client = mongo_client = MongoClient(url)[self._db_name]

    def close(self):
        """ Close mongo connection """
        self._mongo_client.close()

    def get_collection_names(self):
        """ Return names of existing collections in mongo database """
        return self._mongo_client.list_collection_names()

    def find_document(self, collection_name: str, object_id: str, filters: [{'prop': str, 'value': str}] = {}):
        """ Find document with id in parameter, from collection type defined, and using filters if any """
        if type(object_id) is not str:
            raise Exception(f'mongo.find_document : "object_id" must be a string')
        collection = self._mongo_client[collection_name]
        if collection is None:
            raise Exception(f'Mongo error : {collection_name} is not existing in {self._db_name}')

        filters = {**filters, "_id": object_id}

        found_object = list(collection.find(filters))
        if not found_object or len(found_object) == 0:
            return None

        return self._parse_result(list(found_object))[0]

    def find_documents(self, collection_name: str, object_ids: [str] = None, filters: [{'prop': str, 'value': str}] = {}):
        """ Find documents with ids in parameter, and using filters if any
            If no ids are provided, returns all documents for collection type in parameter
        """
        if object_ids and type(object_ids) is not list:
            raise Exception(f'mongo.find_document: "object_ids" must be a list')
        collection = self._mongo_client[collection_name]
        if collection is None:
            raise Exception(f'Mongo error : {collection_name} is not existing in {self._db_name}')

        if object_ids:
            filters = {**filters, "_id": {"$in": object_ids}}

        found_objects = list(collection.find(filters))
        return self._parse_result(found_objects)

    def _parse_result(self, objects: [object]):
        """ Pop _id and __v properties from result, and populate an id field """
        parsed_objects = []
        for obj in objects:
            parsed_object = {**obj, 'id': obj['_id']}
            parsed_object.pop('_id')
            parsed_object.pop('__v', None)
            parsed_objects.append(parsed_object)
        return parsed_objects



import os
import asyncio
import json
import logging
from ..drivers import AMQPClient, GraphDriver, MongoDriver, IndexDriver
from ..datastore import Datastore
from ..tasks import TaskConsumer


async def create_amqp_client(loop):
    """ Create an amqp client ready for connection """
    amqp_url = f"amqp://{os.environ['RABBITMQ_USER']}:{os.environ['RABBITMQ_PWD']}@{os.environ['RABBITMQ_HOST']}:{os.environ['RABBITMQ_PORT']}"
    exchange_name = f"{os.environ['ENV']}-{os.environ['SOLUTION_NAME']}"
    amqp_client = AMQPClient(amqp_url, exchange_name, loop)
    return amqp_client


async def create_datastore(
    loop,
    amqp_client,
    json_schema,
    nodes_namespace_uri,
    nodes_prefix,
    namespace_mapping,
    model_class_mapping=None,
    use_rdf=True,
    use_index=True,
    use_mongo=True
):
    """ Helper for creating a datastore with drivers and schema """

    graph_driver = index_driver = mongo_driver = None

    # Instanciate driver for querying graph triplestore
    if use_rdf:
        graph_driver = GraphDriver(
            amqp_client,
            f"{os.environ['MODULE_ID']}-graph-driver",
            os.environ['RDF_REPOSITORY_NAME'],
            namespace_mapping,
            nodes_namespace_uri,
            nodes_prefix
        )

    # Instanciate driver for querying index
    if use_index:
        index_driver = IndexDriver(
            host=os.environ['ES_HOST'],
            port=os.environ['ES_PORT'],
            user=os.environ['ES_USER'],
            pwd=os.environ['ES_PWD'],
            index_prefix=os.environ['INDEX_PREFIX']
        )

    # Instanciate driver for querying mongo
    if use_mongo:
        mongo_url = f"mongodb://{os.environ['MONGO_USER']}:{os.environ['MONGO_PWD']}@{os.environ['MONGO_ENDPOINT']}"
        mongo_driver = MongoDriver(mongo_url, os.environ['MONGO_DATABASE'])

    # Define datastore with model layer and database drivers
    datastore = Datastore(
        json_schema,
        namespace_mapping,
        nodes_namespace_uri,
        nodes_prefix,
        graph_driver,
        mongo_driver,
        index_driver,
        # mapper for model classes, necessary for using model layer
        model_class_mapper=model_class_mapping
    )

    return datastore



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['elz_py_tools',
 'elz_py_tools.datastore',
 'elz_py_tools.drivers',
 'elz_py_tools.drivers.amqp',
 'elz_py_tools.drivers.graph',
 'elz_py_tools.drivers.index',
 'elz_py_tools.drivers.mongo',
 'elz_py_tools.drivers.toolkit',
 'elz_py_tools.tasks',
 'elz_py_tools.utilities']

package_data = \
{'': ['*']}

install_requires = \
['aio-pika>=6.8.0,<7.0.0',
 'elasticsearch>=7.13.3,<8.0.0',
 'pymongo>=3.11.4,<4.0.0']

setup_kwargs = {
    'name': 'elz-py-tools',
    'version': '0.1.3',
    'description': 'Python components compliant with Synaptix framework',
    'long_description': '# Python tools\n\nRequires Python 3.8+\n\nPyPI : https://pypi.org/project/elz-py-tools/\n\nThis repository contains some Python components used in Elzeard project.\nThese components contain AMQP client and drivers for connecting databases. They are made to be working along with synaptix framework created by Mnemotix (https://gitlab.com/mnemotix).\n\nThis repository uses Poetry for package dependencies (https://python-poetry.org/).\n\nFor working on these components or new ones, install poetry and install package dependencies by running :\n\n> poetry install\n\nFor using it locally from another repository :\n\n> poetry build\n\n> pip install <local_path_to_this_repo>\n\n\n## Datastore and drivers\n\nTODO: to complete...\n\n\n## Model generator\n\nModel files can be generated for facilitating use of this framework and querying of graph and index\nTODO: complete doc...\n\nHere is the format of the schema json file, which is used by datastore :\n```\n[{\n    schema1: {\n        props: {\n            <prop1>: { dataType: <string> }\n        },\n        links: {\n            <link1>: { relatedType: <string>, nested: <boolean>, plural: <boolean> }\n        }\n    },\n    ...\n}]\n```\n\n## TODO:\n\n- Implements GraphDriver.get_nodes method\n- Add possibility to query a label property in a specific language\n- Option for passing custom datatype mapping in datastore',
    'author': 'Hervé',
    'author_email': 'herve.descombe@elzeard.co',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://gitlab.com/elzeard/elz-py-tools',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.8,<4.0',
}


setup(**setup_kwargs)





# TODO: consider using and extending SPARQLBurger (https://pmitzias.com/SPARQLBurger/) ?


class SPARQLBuilder(object):
    """ Wrapper for building some SPARQL queries """

    def __init__(self, prefixes_mapping: dict, nodes_namespace_uri: str, nodes_prefix: str):
        self._prefixes_mapping = prefixes_mapping
        self._nodes_prefix = nodes_prefix
        self._nodes_namespace_uri = nodes_namespace_uri

    def build_construct_object_query(self, node_type: str, object_uri: str):
        """ Build a construct query for getting properties of uri in parameter """ 
        if ':' not in node_type:
            raise Exception('Node type must contain prefix')
        type_prefix, type_suffix = node_type.split(':')
        node_type_uri = f'{self._prefixes_mapping[type_prefix]}{type_suffix}'

        return f'''
{self._build_prefixes()}
CONSTRUCT {{
    <{object_uri}> ?p ?o .
    <{object_uri}> a <{node_type_uri}>
}} where {{
    <{object_uri}> ?p ?o .
    <{object_uri}> a <{node_type_uri}>
}}
'''

    def build_construct_objects_query(self, node_type: str):
        """ Build a construct query for getting nodes and properties of uris of type in parameter """
        if ':' not in node_type:
            raise Exception('Node type must contain prefix')
        type_prefix, type_suffix = node_type.split(':')
        node_type_uri = f'{self._prefixes_mapping[type_prefix]}{type_suffix}'

        return f'''
{self._build_prefixes()}
CONSTRUCT {{
    ?s ?p ?o .
    ?s a <{node_type_uri}>
}} where {{
    ?s ?p ?o .
    ?s a <{node_type_uri}>
}}
    '''

    def _build_prefixes(self):
        """ Build PREFIX part of query """
        return '\n'.join(f'PREFIX {prefix}: <{uri}>' for prefix, uri in self._prefixes_mapping.items())



import json
import os
from pathlib import Path
from ..datastore.model import DATATYPE_MAPPER, KEY_DATATYPE, KEY_RELATED_TYPE, KEY_NESTED, KEY_PLURAL


MODELS_SUBDIR = "models"


def to_snake_case(str):
    return ''.join(['_'+i.lower() if i.isupper() else i for i in str]).lstrip('_')


def generate_model(model_name, model_spec, output_dir, filename):
    """ Generate Python class for model in parameter
        Output file is generated into output_dir and model subdirectory
    """
    props_declarations = []
    links_declarations = []

    # Generate props
    for prop_name, prop_spec in model_spec.get('props', []).items():
        if prop_name == "type":  # reserved word in Python
            prop_name = prop_name + "_"

        target_type = DATATYPE_MAPPER.get(prop_spec[KEY_DATATYPE], None)
        docstring = "Return an array of values of type ".format(related_type=target_type if target_type else "object") \
            if prop_spec.get(KEY_PLURAL, False) \
            else "Return a value of type {related_type}".format(related_type=target_type if target_type else "object")

        props_declarations.append("""
    @property
    def {prop_accessor}(self):
        \"\"\" {docstring} \"\"\"
        return self.get_prop(\"{prop_name}\")
""".format(prop_accessor=to_snake_case(prop_name), docstring=docstring, prop_name=prop_name))

    # Generate links
    for link_name, link_spec in model_spec.get('links', []).items():
        internal_prop_name = link_spec.get('internalProp', None) or link_name
        related_type = link_spec['relatedType']
        docstring = "Return multiple objects of type {related_type}".format(related_type=related_type) \
            if link_spec["plural"] \
            else "Return single object of type {related_type}".format(related_type=related_type)

        links_declarations.append("""
    async def {link_accessor}(self):
        \"\"\" {docstring} \"\"\"
        return await self.get_link(\"{internal_name}\")
""".format(link_accessor=to_snake_case(link_name), docstring=docstring, internal_name=internal_prop_name))

    file_content = """from elz_py_tools.datastore.model import Model


class {class_name}(Model):

    @staticmethod
    def get_schema_type():
        return "{schema_type}"
{props}{links}""".format(
        class_name=model_spec['name'],
        schema_type=model_spec['name'],
        props=''.join(props_declarations),
        links=''.join(links_declarations))

    filepath = "{dir}/{filename}.py".format(dir=output_dir, filename=filename)
    model_file = open(filepath, "w+")
    model_file.write(file_content)
    model_file.close()


def generate_models(json_file: str, output_dir: str):
    """ Generate models contained in json schema file in argument
        Files are generated into output_dir
    """
    if not os.path.exists(json_file):
        raise Exception("json_file path does not exist")

    with open(json_file, 'r+') as f:
        model_schemas = json.load(f)

    output_path = Path(output_dir)
    if not output_path.is_dir():
        raise Exception("output_dir must be a directory or is not existing")

    # Create model subdirectory
    models_dir_path = os.path.join(output_dir, MODELS_SUBDIR)
    if not Path(models_dir_path).is_dir():
        os.mkdir(models_dir_path)

    models_built = []

    # Generate model and build tuples: (filename, classname)
    for model_name, spec in model_schemas.items():
        filename = to_snake_case(model_name)
        if filename == "yield":  # reserved word in Python
            filename = filename + "_model"

        generate_model(model_name, spec, models_dir_path, filename)
        models_built.append((filename, spec['name']))

    # Generate imports of all model classes into an __init__.py
    model_imports = "{imports}".format(imports="\n".join([
        "from .{model_dir}.{filename} import {class_name}".format(
            model_dir=MODELS_SUBDIR,
            filename=to_snake_case(model_filename),
            class_name=class_name)
        for model_filename, class_name in models_built]))

    # Generate mapping of all model classes with model name
    model_mapping = """
model_class_mapper = {{
    {mapping}
}}""".format(mapping=',\n    '.join([
        '"{class_name}": {class_name}'.format(class_name=class_name)
        for _, class_name in models_built]))

    # Generate __init__.py file
    init_file_content = """{imports}

{mapping}
""".format(imports=model_imports, mapping=model_mapping)

    init_filepath = os.path.join(output_dir, "__init__.py")
    model_file = open(init_filepath, "w+")
    model_file.write(init_file_content)
    model_file.close()


if __name__ == "__main__":
    generate_models("../datastore/test.json", "./test")



import logging
from aio_pika import IncomingMessage
from ..drivers import AMQPClient


LOGGER = logging.getLogger(__name__)


class TaskConsumer(object):
    """Wrapper for listening amqp bus and triggering tasks"""

    _amqp_consumer: AMQPClient

    def __init__(self, amqp_client: AMQPClient, routing_keys: [str]):
        self._amqp_consumer = amqp_client
        self._routing_keys = routing_keys
        self._callbacks = {}

    def bind(self, command: str, callback):
        """Bind a callback to command"""
        if command not in self._callbacks:
            self._callbacks[command] = []
        self._callbacks[command].append(callback)

    async def listen(self):
        """Listen on bus, for routing keys defined in contructor"""
        for routing_key in self._routing_keys:
            await self._amqp_consumer.listen(routing_key, self._on_message)

    async def _on_message(self, message: IncomingMessage):
        """Dispatch message to the matching callback"""
        command = message["headers"]["command"]
        if command not in self._callbacks:
            raise Exception(f'Task consumer received unhandled message of type {command}')

        for callback in self._callbacks[command]:
            try:
                response = await callback(message["body"])
            except Exception as e:
                # TODO: send mail to a support address ?
                raise e

        return response
